"""
Managers for our models
"""

from django.db import models


class SettingsManager(models.Manager):
    """
    AFatManager
    """

    def get_setting(self, setting_key: str) -> str:
        """
        Apply select_related for default query optimizations.
        """

        return self.values_list("value", flat=True).get(variable__exact=setting_key)



"""
Models
"""

from django.contrib.auth.models import Group, User
from django.db import models
from django.dispatch import receiver
from django.utils import timezone
from django.utils.text import slugify
from django.utils.translation import gettext as _

from aa_forum.managers import SettingsManager


def get_sentinel_user() -> User:
    """
    Get user or create one
    :return:
    :rtype:
    """

    return User.objects.get_or_create(username="deleted")[0]


def get_slug_on_save(subject: str) -> str:
    """
    Get the slug
    :param subject:
    :type subject:
    :return:
    :rtype:
    """

    run = 0
    subject_slug = slugify(subject, allow_unicode=True)

    while Slugs.objects.filter(slug=subject_slug).exists():
        run += 1
        subject_slug = slugify(subject + "-" + str(run), allow_unicode=True)

    Slugs(slug=subject_slug).save()

    slug_to_return = Slugs.objects.get(slug=subject_slug)

    return slug_to_return


class General(models.Model):
    """
    Meta model for app permissions
    """

    class Meta:
        """
        Meta definitions
        """

        verbose_name = "AA-Forum"
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", _("Can access the AA-Forum module")),
            (
                "manage_forum",
                _("Can manage the AA-Forum module (Categories, topics and messages)"),
            ),
        )


class Slugs(models.Model):
    """
    Slugs
    """

    slug = models.SlugField(max_length=255, allow_unicode=True)

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Slug")
        verbose_name_plural = _("Slugs")

    @receiver(models.signals.post_delete, sender="aa_forum.Categories")
    def handle_deleted_category(sender, instance, **kwargs):
        """
        Delete category slug, when category is deleted
        :param instance:
        :type instance:
        :param kwargs:
        :type kwargs:
        """

        instance.slug.delete()

    @receiver(models.signals.post_delete, sender="aa_forum.Boards")
    def handle_deleted_board(sender, instance, **kwargs):
        """
        Delete board slug, when board is deleted
        :param instance:
        :type instance:
        :param kwargs:
        :type kwargs:
        """

        instance.slug.delete()

    @receiver(models.signals.post_delete, sender="aa_forum.Topics")
    def handle_deleted_topic(sender, instance, **kwargs):
        """
        Delete topic slug, when topic is deleted
        :param instance:
        :type instance:
        :param kwargs:
        :type kwargs:
        """

        instance.slug.delete()

    def __str__(self) -> str:
        return str(self.slug)


class Categories(models.Model):
    """
    Categories
    """

    name = models.CharField(max_length=254, default="")
    slug = models.ForeignKey(
        Slugs,
        blank=True,
        null=True,
        related_name="+",
        on_delete=models.CASCADE,
    )
    is_collapsible = models.BooleanField(default=False)
    order = models.IntegerField(default=0)

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Category")
        verbose_name_plural = _("Categories")

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        """
        Add the slug on save
        :param force_insert:
        :type force_insert:
        :param force_update:
        :type force_update:
        :param using:
        :type using:
        :param update_fields:
        :type update_fields:
        """

        if self.slug is None or self.slug == "":
            slug = get_slug_on_save(subject=self.name)
            self.slug = slug

        super().save()

    def __str__(self) -> str:
        return str(self.name)


class Boards(models.Model):
    """
    Boards
    """

    category = models.ForeignKey(
        Categories,
        related_name="boards",
        on_delete=models.CASCADE,
    )
    name = models.CharField(max_length=254, default="")
    slug = models.ForeignKey(
        Slugs,
        blank=True,
        null=True,
        related_name="+",
        on_delete=models.CASCADE,
    )
    description = models.TextField(null=True, blank=True)
    parent_board = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="child_boards",
        on_delete=models.CASCADE,
    )
    groups = models.ManyToManyField(
        Group,
        blank=True,
        related_name="aa_forum_boards_group_restriction",
    )
    order = models.IntegerField(default=0)

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Board")
        verbose_name_plural = _("Boards")

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        """
        Add the slug on save
        """

        if self.slug is None or self.slug == "":
            slug = get_slug_on_save(subject=self.name)
            self.slug = slug

        super().save()

    def __str__(self) -> str:
        return str(self.name)

    def latest_topic(self):
        """
        Get the latest topic for this board
        :return:
        :rtype:
        """

        topic = Topics.objects.filter(board=self).order_by("time_modified").last()

        return topic


class Topics(models.Model):
    """
    Topics
    """

    subject = models.CharField(max_length=254, default="")
    slug = models.ForeignKey(
        Slugs,
        blank=True,
        null=True,
        related_name="+",
        on_delete=models.CASCADE,
    )
    is_sticky = models.BooleanField(
        default=False,
        db_index=True,
    )
    is_locked = models.BooleanField(
        default=False,
        db_index=True,
    )
    board = models.ForeignKey(
        Boards,
        related_name="topics",
        on_delete=models.CASCADE,
    )
    user_started = models.ForeignKey(
        User,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
    )
    user_updated = models.ForeignKey(
        User,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
    )
    num_views = models.IntegerField(default=0)
    num_replies = models.IntegerField(default=0)
    time_modified = models.DateTimeField(default=timezone.now)
    read_by = models.ManyToManyField(
        User,
        blank=True,
        related_name="aa_forum_read_topics",
    )

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Topic")
        verbose_name_plural = _("Topics")

        ordering = ["-time_modified"]

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        """
        Add the slug on save
        """

        if self.slug is None or self.slug == "":
            slug = get_slug_on_save(subject=self.subject)
            self.slug = slug

        super().save()

    def first_message(self):
        """
        Get the first message for this topic
        :return:
        :rtype:
        """

        message = Messages.objects.filter(topic=self).first()

        return message

    def last_message(self):
        """
        Get the latest message for this topic
        :return:
        :rtype:
        """

        message = Messages.objects.filter(topic=self).last()

        return message


class Messages(models.Model):
    """
    Messages
    """

    board = models.ForeignKey(
        Boards,
        related_name="messages",
        on_delete=models.CASCADE,
    )
    topic = models.ForeignKey(
        Topics,
        related_name="messages",
        on_delete=models.CASCADE,
    )
    time_posted = models.DateTimeField(default=timezone.now)
    time_modified = models.DateTimeField(default=timezone.now)
    user_created = models.ForeignKey(
        User,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
    )
    user_updated = models.ForeignKey(
        User,
        blank=True,
        null=True,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
    )
    message = models.TextField(null=True, blank=True)
    read_by = models.ManyToManyField(
        User,
        blank=True,
        related_name="aa_forum_read_messages",
    )

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Message")
        verbose_name_plural = _("Messages")

    def __str__(self) -> str:
        return str(self.pk)


class PersonalMessages(models.Model):
    """
    Personal messages
    """

    sender = models.ForeignKey(
        User,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
    )
    recipient = models.ForeignKey(
        User,
        related_name="+",
        on_delete=models.SET(get_sentinel_user),
    )
    time_sent = models.DateTimeField(default=timezone.now)
    subject = models.CharField(max_length=254, default="")
    slug = models.ForeignKey(
        Slugs,
        blank=True,
        null=True,
        related_name="+",
        on_delete=models.CASCADE,
    )
    message = models.TextField(null=True, blank=True)
    is_read = models.BooleanField(default=False)

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Personal Message")
        verbose_name_plural = _("Personal Messages")

    def save(
        self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        """
        Add the slug on save
        """

        if self.slug is None or self.slug == "":
            slug = get_slug_on_save(subject=self.subject)
            self.slug = slug

        super().save()

    def __str__(self) -> str:
        return str(self.subject)


class Settings(models.Model):
    """
    Settings
    """

    variable = models.CharField(
        max_length=254, blank=False, primary_key=True, unique=True
    )
    value = models.TextField(blank=False)

    objects = SettingsManager()

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Setting")
        verbose_name_plural = _("Settings")



"""
Django admin declarations
"""

from django.contrib import admin

from aa_forum.models import Boards, Categories


@admin.register(Categories)
class CategoriesAdmin(admin.ModelAdmin):
    """
    Categories admin
    """

    list_display = ("name", "slug")
    exclude = ("slug", "is_collapsible", "order")


@admin.register(Boards)
class BoardsAdmin(admin.ModelAdmin):
    """
    Boards admin
    """

    list_display = ("name", "slug")
    exclude = ("slug", "parent_board", "groups", "order")



# Generated by Django 3.2.4 on 2021-06-09 13:10

from django.db import migrations

default_settings_to_migrate = [
    {"variable": "defaultDateFormat", "value": "F d, Y"},
    {"variable": "defaultTimeFormat", "value": "G:i:s"},
    {"variable": "maxMessageLength", "value": "20000"},
    {"variable": "defaultMaxMessages", "value": "15"},
    {"variable": "defaultMaxTopics", "value": "20"},
    {"variable": "oldTopicDays", "value": "120"},
]


def add_default_setting(apps, schema_editor):
    """
    Add default settings on migration
    :param apps:
    :type apps:
    :param schema_editor:
    :type schema_editor:
    """

    Settings = apps.get_model("aa_forum", "Settings")
    db_alias = schema_editor.connection.alias

    default_settings = []
    for default_setting in default_settings_to_migrate:
        default_settings.append(
            Settings(
                variable=default_setting["variable"], value=default_setting["value"]
            )
        )

    Settings.objects.using(db_alias).bulk_create(default_settings)


def remove_default_settings(apps, schema_editor):
    """
    Remove default settings on migration to zero
    :param apps:
    :type apps:
    :param schema_editor:
    :type schema_editor:
    """

    Settings = apps.get_model("aa_forum", "Settings")
    db_alias = schema_editor.connection.alias

    for default_setting in default_settings_to_migrate:
        Settings.objects.using(db_alias).filter(
            variable=default_setting["variable"]
        ).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("aa_forum", "0002_reserved_slugs"),
    ]

    operations = [
        migrations.RunPython(add_default_setting, remove_default_settings),
    ]



"""
Administration related views
"""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Prefetch
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.shortcuts import redirect, render
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from aa_forum.forms import EditBoardForm, EditCategoryForm
from aa_forum.models import Boards, Categories


@login_required
@permission_required("aa_forum.manage_forum")
def index(request: WSGIRequest) -> HttpResponse:
    """
    Administration index view
    :param request:
    :type request:
    :return:
    :rtype:
    """

    categories = (
        Categories.objects.prefetch_related(
            Prefetch(
                "boards",
                queryset=Boards.objects.filter(parent_board__isnull=True).order_by(
                    "order"
                ),
            )
        )
        .all()
        .order_by("order")
    )

    category_loop = list()
    for category in categories:
        boards_data = [
            {
                "board_obj": board,
                "board_edit_form": EditBoardForm(
                    prefix="edit-board-" + str(board.id), instance=board
                ),
            }
            for board in category.boards.all()
        ]
        category_data = {
            "category_obj": category,
            "category_forms": {
                "new_board": EditBoardForm(
                    prefix="new-board-in-category-" + str(category.id)
                ),
                "edit_category": EditCategoryForm(
                    prefix="edit-category-" + str(category.id), instance=category
                ),
            },
            "boards": boards_data,
        }

        category_loop.append(category_data)

    form_new_category = EditCategoryForm(prefix="new-category")

    context = {
        "new_category_form": form_new_category,
        "category_loop": category_loop,
    }

    return render(request, "aa_forum/view/administration/index.html", context)


@login_required
@permission_required("aa_forum.manage_forum")
def category_create(request: WSGIRequest) -> HttpResponseRedirect:
    """
    Create a new category
    :param request:
    :type request:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditCategoryForm(request.POST, prefix="new-category")

        # Check whether it's valid:
        if form.is_valid():
            new_category = Categories()
            new_category.name = form.cleaned_data["name"]
            new_category.order = 999999
            new_category.save()

            messages.success(
                request,
                mark_safe(_("<h4>Success!</h4><p>Category created.</p>")),
            )

    return redirect("aa_forum:admin_index")


@login_required
@permission_required("aa_forum.manage_forum")
def category_edit(request: WSGIRequest, category_id: int) -> HttpResponseRedirect:
    """
    Edit a category
    :param request:
    :type request:
    :param category_id:
    :type category_id:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditCategoryForm(
            request.POST, prefix="edit-category-" + str(category_id)
        )

        if form.is_valid():
            category = Categories.objects.get(pk=category_id)
            category.name = form.cleaned_data["name"]
            category.save()

            messages.success(
                request,
                mark_safe(_("<h4>Success!</h4><p>Category changed.</p>")),
            )

    return redirect("aa_forum:admin_index")


@login_required
@permission_required("aa_forum.manage_forum")
def category_delete(request: WSGIRequest, category_id: int) -> HttpResponseRedirect:
    """
    Edit a board
    :param request:
    :type request:
    :param category_id:
    :type category_id:
    """

    Categories.objects.get(pk=category_id).delete()

    messages.success(
        request,
        mark_safe(_("<h4>Success!</h4><p>Category removed.</p>")),
    )

    return redirect("aa_forum:admin_index")


@login_required
@permission_required("aa_forum.manage_forum")
def board_create(request: WSGIRequest, category_id: int) -> HttpResponseRedirect:
    """
    Create a new board
    :param request:
    :type request:
    :param category_id:
    :type category_id:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditBoardForm(
            request.POST, prefix="new-board-in-category-" + str(category_id)
        )
        board_category = Categories.objects.get(pk=category_id)

        # Check whether it's valid:
        if form.is_valid():
            new_board = Boards()
            new_board.name = form.cleaned_data["name"]
            new_board.description = form.cleaned_data["description"]
            new_board.category = board_category
            new_board.order = 999999
            new_board.save()

            new_board.groups.set(form.cleaned_data["groups"])

            messages.success(
                request,
                mark_safe(_("<h4>Success!</h4><p>Board created.</p>")),
            )

    return redirect("aa_forum:admin_index")


@login_required
@permission_required("aa_forum.manage_forum")
def board_edit(
    request: WSGIRequest, category_id: int, board_id: int
) -> HttpResponseRedirect:
    """
    Edit a board
    :param request:
    :type request:
    :param category_id:
    :type category_id:
    :param board_id:
    :type board_id:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditBoardForm(request.POST, prefix="edit-board-" + str(board_id))

        # Check whether it's valid:
        if form.is_valid():
            board = Boards.objects.get(pk=board_id, category_id=category_id)
            board.name = form.cleaned_data["name"]
            board.description = form.cleaned_data["description"]
            board.groups.set(form.cleaned_data["groups"])
            board.save()

            messages.success(
                request,
                mark_safe(_("<h4>Success!</h4><p>Board changed.</p>")),
            )

    return redirect("aa_forum:admin_index")


@login_required
@permission_required("aa_forum.manage_forum")
def board_delete(
    request: WSGIRequest, category_id: int, board_id: int
) -> HttpResponseRedirect:
    """
    Edit a board
    :param request:
    :type request:
    :param category_id:
    :type category_id:
    :param board_id:
    :type board_id:
    """

    Boards.objects.get(pk=board_id, category_id=category_id).delete()

    messages.success(
        request,
        mark_safe(_("<h4>Success!</h4><p>Board removed.</p>")),
    )

    return redirect("aa_forum:admin_index")


@login_required
@permission_required("aa_forum.manage_forum")
def ajax_category_order(request: WSGIRequest) -> JsonResponse:
    """
    Ajax call :: Save the category order
    :param request:
    :type request:
    """

    data = list()

    if request.method == "POST":
        import simplejson

        categories = simplejson.loads(request.POST.get("categories"))

        for category in categories:
            Categories.objects.update_or_create(
                pk=category["catId"],
                defaults={"order": category["catOrder"]},
            )

        data.append({"success": True})

    return JsonResponse(data, safe=False)


@login_required
@permission_required("aa_forum.manage_forum")
def ajax_board_order(request: WSGIRequest) -> JsonResponse:
    """
    Ajax call :: Save the board order
    :param request:
    :type request:
    """

    data = list()

    if request.method == "POST":
        import simplejson

        boards = simplejson.loads(request.POST.get("boards"))

        for board in boards:
            Boards.objects.update_or_create(
                pk=board["boardId"],
                defaults={"order": board["boardOrder"]},
            )

        data.append({"success": True})

    return JsonResponse(data, safe=False)
